/* SCPI page */

.text-view-scpi {
  background-color: shade(@theme_base_color, .90);
  border-color: alpha(black, 0.8);
  border-top-left-radius: 5px;
  border-top-right-radius: 5px;
  padding: 5px;
}

.text-view-scpi text {
  background-color: rgba(0,0,0,0);
}

.entry-scpi {
  background-color: shade(@theme_base_color, .80);
  border: solid @borders 0px;
  border-top-width: 1px;
  border-color: @theme_bg_color;
  border-top-color: alpha(black, 0.2);
  border-top-left-radius: 0px;
  border-top-right-radius: 0px;
  border-bottom-left-radius: 5px;
  border-bottom-right-radius: 5px;
  outline-width: 0px;
}

.scpi-label {
  font-weight: bolder;
  font-size: 0.7em;
  color: alpha(@theme_text_color, 0.3);
}

#button-scpi-clear {
  background-color: @warning_bg_color;
}


/* Screenshot page */



/* Benchmark page */

.page-benchmark trough {
  min-height: 60px;
}

.page-benchmark progress {
  min-height: 60px;
}


.label-benchmark-result {
  font-weight: bold;
  font-size: 1.8em;
}



/* Script page */

.box-script-toolbar {
  margin-top: 10px;
}

.source-view-script {
  border-top-left-radius: 5px;
  border-top-right-radius: 5px;
  padding: 10px;
}

.text-view-script-status {
  background-color: shade(@theme_base_color, .80);
  border-bottom-left-radius: 5px;
  border-bottom-right-radius: 5px;
  padding: 10px;
  border-top-width: 1px;
  border-top-color: alpha(black, 0.2);
}

.text-view-script-status text {
  background-color: rgba(0,0,0,0);
}
